import json
from typing import Dict, List, Optional, Tuple
from uuid import uuid4, UUID
from promptmodel.database.models import (
    PromptModel,
    PromptModelVersion,
    Prompt,
    RunLog,
    SampleInputs,
    DeployedPromptModel,
    DeployedPromptModelVersion,
    DeployedPrompt,
)
from playhouse.shortcuts import model_to_dict
from promptmodel.utils.enums import PromptModelVersionStatus, ParsingType
from promptmodel.utils.random_utils import select_version
from promptmodel.utils import logger
from promptmodel.database.config import db


# Insert
def create_prompt_model(name: str, project_uuid: str):
    """Create a new PromptModel with the given name."""
    return PromptModel.create(uuid=uuid4(), name=name, project_uuid=project_uuid)


def create_prompt_models(prompt_model_list: list):
    """Create PromptModels with List of Dict"""
    with db.atomic():
        PromptModel.insert_many(prompt_model_list).execute()
    return


def create_prompt_model_version(
    prompt_model_uuid: str,
    from_uuid: Optional[str],
    status: str,
    model: str,
    parsing_type: Optional[ParsingType] = None,
    output_keys: Optional[List[str]] = None,
    functions: List[str] = [],
):
    """Create a new PromptModel version with the given parameters."""
    return PromptModelVersion.create(
        uuid=uuid4(),
        from_uuid=from_uuid,
        prompt_model_uuid=prompt_model_uuid,
        status=status,
        model=model,
        parsing_type=parsing_type,
        output_keys=output_keys,
        functions=functions,
    )


def create_prompt_model_versions(prompt_model_version_list: list):
    """Create PromptModel versions with List of Dict"""
    with db.atomic():
        PromptModelVersion.insert_many(prompt_model_version_list).execute()
    return


def create_prompt(
    version_uuid: str,
    role: str,
    step: int,
    content: str,
):
    """Create a new prompt with the given parameters."""
    return Prompt.create(
        version_uuid=version_uuid,
        role=role,
        step=step,
        content=content,
    )


def create_prompts(prompt_list: list):
    """Create prompts with List of Dict"""
    with db.atomic():
        Prompt.insert_many(prompt_list).execute()
    return


def create_run_log(
    prompt_model_version_uuid: str,
    inputs: str,
    raw_output: str,
    parsed_outputs: str,
    is_deployed: bool = False,
    function_call: Optional[dict] = None,
):
    """Create a new run log with the given parameters."""
    return RunLog.create(
        version_uuid=prompt_model_version_uuid,
        inputs=inputs,
        raw_output=raw_output,
        parsed_outputs=parsed_outputs,
        is_deployed=is_deployed,
        function_call=function_call,
    )


def create_run_logs(run_log_list: list):
    """Create run_logs with List of Dict"""
    with db.atomic():
        RunLog.insert_many(run_log_list).execute()
    return


# Update For delayed Insert
def update_prompt_model_version(prompt_model_version_uuid: str, status: str):
    """Update the status of the given PromptModel version."""
    return (
        PromptModelVersion.update(status=status)
        .where(PromptModelVersion.uuid == prompt_model_version_uuid)
        .execute()
    )


# Select all
def list_prompt_models() -> List[Dict]:
    """List all PromptModels."""
    response: List[PromptModel] = list(PromptModel.select())
    return [model_to_dict(x, recurse=False) for x in response]


def list_prompt_model_versions(prompt_model_uuid: str) -> List[Dict]:
    """List all PromptModel versions for the given PromptModel."""
    response: List[PromptModelVersion] = list(
        PromptModelVersion.select()
        .where(PromptModelVersion.prompt_model_uuid == prompt_model_uuid)
        .order_by(PromptModelVersion.created_at)
    )
    return [model_to_dict(x, recurse=False) for x in response]


def list_prompts(prompt_model_version_uuid: str) -> List[Dict]:
    """List all prompts for the given PromptModel version."""
    response: List[Prompt] = list(
        Prompt.select()
        .where(Prompt.version_uuid == prompt_model_version_uuid)
        .order_by(Prompt.step)
    )
    return [model_to_dict(x, recurse=False) for x in response]


def list_samples():
    """List all samples"""
    response = list(SampleInputs.select())
    return [x.__data__ for x in response]


def list_run_logs(prompt_model_version_uuid: str) -> List[RunLog]:
    """List all run logs for the given PromptModel version."""
    response: List[RunLog] = list(
        RunLog.select()
        .where(RunLog.version_uuid == prompt_model_version_uuid)
        .order_by(RunLog.created_at.desc())
    )
    return [model_to_dict(x, recurse=False) for x in response]


# Select one
def get_prompt_model_uuid(prompt_model_name: str) -> Dict:
    """Get uuid of prompt_model by name"""
    try:
        response = PromptModel.get(PromptModel.name == prompt_model_name)
        return model_to_dict(response, recurse=False)
    except:
        return None


def get_sample_input(sample_name: str) -> Dict:
    """Get sample input from local DB"""
    try:
        response = SampleInputs.get(SampleInputs.name == sample_name)
        sample = model_to_dict(response, recurse=False)
        # sample["contents"] = json.loads(sample["contents"])
        return sample
    except:
        return None


def get_latest_version_prompts(prompt_model_name: str) -> Tuple[List[Prompt], str]:
    try:
        with db.atomic():
            latest_run_log: RunLog = (
                RunLog.select()
                .join(PromptModelVersion)
                .where(
                    PromptModelVersion.prompt_model_uuid
                    == PromptModel.get(PromptModel.name == prompt_model_name).uuid
                )
                .order_by(RunLog.created_at.desc())
                .get()
            )

            prompts: List[Prompt] = (
                Prompt.select()
                .where(Prompt.version_uuid == latest_run_log.version_uuid.uuid)
                .order_by(Prompt.step.asc())
            )

            version: PromptModelVersion = (
                PromptModelVersion.select()
                .where(PromptModelVersion.uuid == latest_run_log.version_uuid.uuid)
                .get()
            )

        version_details = {
            "model": version.model,
            "uuid": version.uuid,
            "parsing_type": version.parsing_type,
            "output_keys": version.output_keys,
        }

        return prompts, version_details

    except Exception as e:
        logger.error(e)
        return None, None


def get_deployed_prompts(prompt_model_name: str) -> Tuple[List[DeployedPrompt], str]:
    try:
        with db.atomic():
            versions: List[DeployedPromptModelVersion] = list(
                DeployedPromptModelVersion.select()
                .join(DeployedPromptModel)
                .where(
                    DeployedPromptModelVersion.prompt_model_uuid
                    == DeployedPromptModel.get(
                        DeployedPromptModel.name == prompt_model_name
                    ).uuid
                )
            )
            prompts: List[DeployedPrompt] = list(
                DeployedPrompt.select()
                .where(
                    DeployedPrompt.version_uuid.in_(
                        [version.uuid for version in versions]
                    )
                )
                .order_by(DeployedPrompt.step.asc())
            )
        # select version by ratio
        selected_version = select_version([version.__data__ for version in versions])
        selected_prompts = list(
            filter(
                lambda prompt: str(prompt.version_uuid.uuid)
                == str(selected_version["uuid"]),
                prompts,
            )
        )

        version_details = {
            "model": selected_version["model"],
            "uuid": selected_version["uuid"],
            "parsing_type": selected_version["parsing_type"],
            "output_keys": selected_version["output_keys"],
        }

        return selected_prompts, version_details
    except Exception as e:
        logger.error(e)
        return None, None


# Update
def update_is_deployed_prompt_model(prompt_model_uuid: str, is_deployed: bool):
    """Update the name of the given PromptModel."""
    return (
        PromptModel.update(is_deployed=is_deployed)
        .where(PromptModel.uuid == prompt_model_uuid)
        .execute()
    )


def update_used_in_code_prompt_model(prompt_model_uuid: str, used_in_code: bool):
    """Update the name of the given PromptModel."""
    return (
        PromptModel.update(used_in_code=used_in_code)
        .where(PromptModel.uuid == prompt_model_uuid)
        .execute()
    )


def update_used_in_code_prompt_model_by_name(
    prompt_model_name: str, used_in_code: bool
):
    """Update the name of the given PromptModel."""
    return (
        PromptModel.update(used_in_code=used_in_code)
        .where(PromptModel.name == prompt_model_name)
        .execute()
    )


def rename_prompt_model(prompt_model_uuid: str, new_name: str):
    """Update the name of the given PromptModel."""
    return (
        PromptModel.update(name=new_name)
        .where(PromptModel.uuid == prompt_model_uuid)
        .execute()
    )


def hide_prompt_model_not_in_code(local_prompt_model_list: list):
    return (
        PromptModel.update(used_in_code=False)
        .where(PromptModel.name.not_in(local_prompt_model_list))
        .execute()
    )


async def update_deployed_cache(project_status: dict):
    """Update Deployed Prompts Cache"""
    # TODO: 효율적으로 수정
    # 현재는 delete all & insert all
    prompt_models = project_status["prompt_models"]
    prompt_model_versions = project_status["prompt_model_versions"]
    for version in prompt_model_versions:
        if version["is_published"] is True:
            version["ratio"] = 1.0
    prompts = project_status["prompts"]

    with db.atomic():
        DeployedPromptModel.delete().execute()
        DeployedPromptModel.insert_many(prompt_models).execute()
        DeployedPromptModelVersion.insert_many(prompt_model_versions).execute()
        DeployedPrompt.insert_many(prompts).execute()
    return


def update_samples(samples: List[Dict]):
    """Update samples"""

    with db.atomic():
        SampleInputs.delete().execute()
        SampleInputs.insert_many(samples).execute()
    return


def update_prompt_model_uuid(local_uuid, new_uuid):
    """Update prompt_model_uuid"""
    with db.atomic():
        local_prompt_model: PromptModel = PromptModel.get(
            PromptModel.uuid == local_uuid
        )
        PromptModel.create(
            uuid=new_uuid,
            name=local_prompt_model.name,
            project_uuid=local_prompt_model.project_uuid,
            created_at=local_prompt_model.created_at,
            used_in_code=local_prompt_model.used_in_code,
            is_deployed=True,
        )
        PromptModelVersion.update(prompt_model_uuid=new_uuid).where(
            PromptModelVersion.prompt_model_uuid == local_uuid
        ).execute()
        PromptModel.delete().where(PromptModel.uuid == local_uuid).execute()
    return


def find_ancestor_version(
    prompt_model_version_uuid: str, versions: Optional[list] = None
):
    """Find ancestor version"""

    # get all versions
    if versions is None:
        response = list(PromptModelVersion.select())
        versions = [model_to_dict(x, recurse=False) for x in response]

    # find target version
    target = list(
        filter(lambda version: version["uuid"] == prompt_model_version_uuid, versions)
    )[0]

    target = _find_ancestor(target, versions)

    prompts = list(Prompt.select().where(Prompt.version_uuid == target["uuid"]))
    prompts = [model_to_dict(x, recurse=False) for x in prompts]
    return target, prompts


def find_ancestor_versions(target_prompt_model_uuid: Optional[str] = None):
    """find ancestor versions for each versions in input"""
    # get all versions
    if target_prompt_model_uuid is not None:
        response = list(
            PromptModelVersion.select().where(
                PromptModelVersion.prompt_model_uuid == target_prompt_model_uuid
            )
        )
    else:
        response = list(PromptModelVersion.select())
    versions = [model_to_dict(x, recurse=False) for x in response]

    targets = list(
        filter(
            lambda version: version["status"]
            == PromptModelVersionStatus.CANDIDATE.value
            and version["version"] is None,
            versions,
        )
    )

    target_and_prompts = [
        find_ancestor_version(target["uuid"], versions) for target in targets
    ]
    targets_with_real_ancestor = [
        target_and_prompt[0] for target_and_prompt in target_and_prompts
    ]
    target_prompts = []
    for target_and_prompt in target_and_prompts:
        target_prompts += target_and_prompt[1]

    return targets_with_real_ancestor, target_prompts


def _find_ancestor(target: dict, versions: List[Dict]):
    ancestor = None
    temp = target
    if target["from_uuid"] is None:
        ancestor = None
    else:
        while temp["from_uuid"] is not None:
            new_temp = [
                version for version in versions if version["uuid"] == temp["from_uuid"]
            ][0]
            if new_temp["version"] is not None:
                ancestor = new_temp
                break
            else:
                temp = new_temp
        target["from_uuid"] = ancestor["uuid"] if ancestor is not None else None

    return target


def update_candidate_prompt_model_version(new_candidates: dict):
    """Update candidate version"""
    with db.atomic():
        for uuid, version in new_candidates.items():
            (
                PromptModelVersion.update(version=version, is_deployed=True)
                .where(PromptModelVersion.uuid == uuid)
                .execute()
            )
        # Find PromptModel
        prompt_model_versions: List[PromptModelVersion] = list(
            PromptModelVersion.select().where(
                PromptModelVersion.uuid.in_(list(new_candidates.keys()))
            )
        )
        prompt_model_uuids = [
            prompt_model.prompt_model_uuid.uuid
            for prompt_model in prompt_model_versions
        ]
        PromptModel.update(is_deployed=True).where(
            PromptModel.uuid.in_(prompt_model_uuids)
        ).execute()



from __future__ import annotations
import asyncio

from dataclasses import dataclass
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

import promptmodel.utils.logger as logger
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.utils.prompt_util import fetch_prompts, run_async_in_sync
from promptmodel.utils.types import LLMStreamResponse, LLMResponse
from promptmodel import DevClient


@dataclass
class PromptModelInterface:
    name: str
    default_model: str


class RegisteringMeta(type):
    def __call__(cls, *args, **kwargs):
        instance: "PromptModel" = super().__call__(*args, **kwargs)
        # Find the global client instance in the current context
        client: Optional[DevClient] = cls.find_client_instance()
        if client is not None:
            client.register_prompt_model(instance.name)
        return instance

    @staticmethod
    def find_client_instance():
        import sys

        # Get the current frame
        frame = sys._getframe(2)
        # Get global variables in the current frame
        global_vars = frame.f_globals
        # Find an instance of DevClient among global variables
        for var_name, var_val in global_vars.items():
            if isinstance(var_val, DevClient):
                return var_val
        return None


class PromptModel(metaclass=RegisteringMeta):
    def __init__(self, name, rate_limit_manager=None):
        self.name = name
        self.llm_proxy = LLMProxy(name, rate_limit_manager)

    def get_prompts(self) -> List[Dict[str, str]]:
        """Get prompt for the promptmodel.
        If dev mode is running(if .promptmodel/config['dev_branch']['online'] = True), it will fetch the latest tested prompt in the dev branch local DB.
        If dev mode is not running, it will fetch the published prompt from the Cloud. (It will be saved in cache DB, so there is no extra latency for API call.)
        - If you made A/B testing in Web Dashboard, it will fetch the prompt randomly by the A/B testing ratio.
        If dev mode is initializing, it will return {}.

        Returns:
            List[Dict[str, str]]: list of prompts. Each prompt is a dict with 'role' and 'content'.
        """
        # add name to the list of prompt_models

        prompts, _ = run_async_in_sync(fetch_prompts(self.name))
        return prompts

    def run(
        self,
        inputs: Dict[str, Any] = {},
        function_list: Optional[List[Dict[str, Any]]] = None,
    ) -> LLMResponse:
        """Run PromptModel. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): input to the promptmodel. Defaults to {}.

        Returns:
            LLMResponse: response from the promptmodel. you can find raw output in response.raw_output or response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        return self.llm_proxy.run(inputs, function_list)

    async def arun(
        self,
        inputs: Dict[str, Any] = {},
        function_list: Optional[List[Dict[str, Any]]] = None,
    ) -> LLMResponse:
        """Async run PromptModel. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): input to the promptmodel. Defaults to {}.

        Returns:
            LLMResponse: response from the promptmodel. you can find raw output in response.raw_output or response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        return await self.llm_proxy.arun(inputs, function_list)

    def stream(
        self,
        inputs: Dict[str, Any] = {},
        function_list: Optional[List[Dict[str, Any]]] = None,
    ) -> Generator[LLMStreamResponse, None, None]:
        """Run PromptModel with stream=True. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): _description_. Defaults to {}.

        Yields:
            Generator[LLMStreamResponse, None, None]: Generator of LLMStreamResponse. you can find raw output in response.raw_output or response.api_response['choices'][0]['delta']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        for item in self.llm_proxy.stream(inputs, function_list):
            yield item

    async def astream(
        self,
        inputs: Optional[Dict[str, Any]] = {},
        function_list: Optional[List[Dict[str, Any]]] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        """Async Run PromptModel with stream=True. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): _description_. Defaults to {}.

        Yields:
            AsyncGenerator[LLMStreamResponse, None]: Generator of LLMStreamResponse. you can find raw output in response.raw_output or response.api_response['choices'][0]['delta']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        async for item in self.llm_proxy.astream(inputs, function_list):
            yield item

    def run_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        function_list: Optional[List[Dict[str, Any]]] = None,
    ) -> LLMResponse:
        """Run PromptModel and make parsed outputs. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): input to the promptmodel. Defaults to {}.

        Returns:
            LLMResponse: response from the promptmodel. you can find parsed outputs in response.parsed_outputs. You can also find raw output in response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        return self.llm_proxy.run_and_parse(inputs, function_list)

    async def arun_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        function_list: Optional[List[Dict[str, Any]]] = None,
    ) -> LLMResponse:
        """Async Run PromptModel and make parsed outputs. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): input to the promptmodel. Defaults to {}.

        Returns:
            LLMResponse: response from the promptmodel. you can find parsed outputs in response.parsed_outputs. You can also find raw output in response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        return await self.llm_proxy.arun_and_parse(inputs, function_list)

    def stream_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        function_list: Optional[List[Dict[str, Any]]] = None,
    ) -> Generator[LLMStreamResponse, None, None]:
        """Run PromptModel with stream=True and make parsed outputs. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): _description_. Defaults to {}.

        Yields:
            Generator[LLMStreamResponse, None, None]: Generator of LLMStreamResponse. you can find parsed outputs in response.parsed_outputs. You can also find raw output in reaponse.raw_output.

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log
        """
        for item in self.llm_proxy.stream_and_parse(inputs, function_list):
            yield item

    async def astream_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        function_list: Optional[List[Dict[str, Any]]] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        """Async Run PromptModel with stream=True and make parsed outputs. It does not raise error.

        Args:
            inputs (Dict[str, Any], optional): _description_. Defaults to {}.

        Yields:
            AsyncGenerator[LLMStreamResponse, None]: Generator of LLMStreamResponse. you can find parsed outputs in response.parsed_outputs. You can also find raw output in reaponse.raw_output.

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log
        """
        async for item in self.llm_proxy.astream_and_parse(inputs, function_list):
            yield item

    # def run_and_parse_function_call(
    #     self,
    #     inputs: Dict[str, Any] = {},
    #     function_list: List[Callable[..., Any]] = [],
    # ) -> Generator[str, None, None]:
    #     return self.llm_proxy.run_and_parse_function_call(inputs, function_list)

    # async def arun_and_parse_function_call(
    #     self,
    #     inputs: Dict[str, Any] = {},
    #     function_list: List[Callable[..., Any]] = [],
    # ) -> Generator[str, None, None]:
    #     return await self.llm_proxy.arun_and_parse_function_call(inputs, function_list)

    # async def astream_and_parse_function_call(
    #     self,
    #     inputs: Dict[str, Any] = {},
    #     function_list: List[Callable[..., Any]] = [],
    # ) -> AsyncGenerator[str, None]:
    #     async for item in self.llm_proxy.astream_and_parse_function_call(
    #         inputs, function_list
    #     ):
    #         yield item



"""This single file is needed to build the Client development dashboard."""

from promptmodel import DevApp
from main import client as main_client

# Example imports
# from <dirname> import < objectname>

app = DevApp()

# Example usage
# This is needed to integrate your codebase with the prompt engineering dashboard

app.include_client(main_client)

app.register_sample(
    name="function_call_test/1",
    content={"user_message": "What is the weather like in Boston?"},
)

from main import get_current_weather, get_current_weather_desc

app.register_function(get_current_weather_desc, get_current_weather)



# test.py
from promptmodel import PromptModel, DevClient
import promptmodel

promptmodel.init()

client = DevClient()

# prompt = PromptModel("function_call_test").get_prompts()

from typing import Optional


def get_current_weather(location: str, unit: Optional[str] = None):
    return "13 degrees celsius"


get_current_weather_desc = {
    "name": "get_current_weather",
    "description": "Get the current weather in a given location",
    "parameters": {
        "type": "object",
        "properties": {
            "location": {
                "type": "string",
                "description": "The city and state, e.g. San Francisco, CA",
            },
            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
        },
        "required": ["location"],
    },
}

import json
import asyncio


@client.register
def main():
    res = PromptModel("function_call_test").run(
        {"user_message": "What is the weather like in Boston?"},
        function_list=[get_current_weather_desc],
    )
    print(res.__dict__)
    if res.function_call:
        print(res.function_call)
        print(get_current_weather(**json.loads(res.function_call["arguments"])))


if __name__ == "__main__":
    main()
    # async def sleep_func():
    #     print(asyncio.all_tasks(loop=None))
    #     asyncio.sleep(100)

    # asyncio.run(sleep_func())



from abc import ABC, abstractmethod
from typing import Callable


class Registry(ABC):
    @abstractmethod
    def register(self, name: str, description: str, func: Callable):
        pass

    # @abstractmethod
    def get(self, name: str):
        pass

    @abstractmethod
    def list(self):
        pass



from enum import Enum


class LocalTask(str, Enum):
    RUN_PROMPT_MODEL = "RUN_PROMPT_MODEL"
    LIST_PROMPT_MODELS = "LIST_PROMPT_MODELS"
    LIST_PROMPT_MODEL_VERSIONS = "LIST_PROMPT_MODEL_VERSIONS"
    LIST_SAMPLES = "LIST_SAMPLES"
    LIST_FUNCTIONS = "LIST_FUNCTIONS"
    GET_PROMPTS = "GET_PROMPTS"
    GET_RUN_LOGS = "GET_RUN_LOGS"
    CHANGE_PROMPT_MODEL_VERSION_STATUS = "CHANGE_PROMPT_MODEL_VERSION_STATUS"
    GET_PROMPT_MODEL_VERSION_TO_SAVE = "GET_PROMPT_MODEL_VERSION_TO_SAVE"
    GET_PROMPT_MODEL_VERSIONS_TO_SAVE = "GET_PROMPT_MODEL_VERSIONS_TO_SAVE"
    UPDATE_CANDIDATE_PROMPT_MODEL_VERSION_ID = (
        "UPDATE_CANDIDATE_PROMPT_MODEL_VERSION_ID"
    )


class ServerTask(str, Enum):
    UPDATE_RESULT_RUN = "UPDATE_RESULT_RUN"
    LOCAL_UPDATE_ALERT = "LOCAL_UPDATE_ALERT"
    UPDATE_RESULT_EVAL = "UPDATE_RESULT_EVAL"


class PromptModelVersionStatus(Enum):
    BROKEN = "broken"
    WORKING = "working"
    CANDIDATE = "candidate"


class ChangeLogAction(str, Enum):
    ADD: str = "ADD"
    DELETE: str = "DEL"
    CHANGE: str = "CHG"
    FIX: str = "FIX"


class Role(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


class ParsingType(str, Enum):
    COLON = "colon"
    SQUARE_BRACKET = "square_bracket"
    DOUBLE_SQUARE_BRACKET = "double_square_bracket"
    HTML = "html"


class ParsingPattern(dict, Enum):
    COLON = {
        "start": r"(.*?): \n",
        "start_fstring": "{key}: \n",
        "end_fstring": None,
        "whole": r"(.*?): (.*?)\n",
        "start_token": None,
        "end_token": None,
    }
    SQUARE_BRACKET = {
        "start": r"\[(\w+)\s+type=([\w,\s\[\]]+)\]",
        "start_fstring": "[{key} type={type}]",
        "end_fstring": "[/{key}]",
        "whole": r"\[(\w+)\s+type=([\w,\s\[\]]+)\](.*?)\[/\1\]",
        "start_token": r"[",
        "end_token": r"]",
    }
    DOUBLE_SQUARE_BRACKET = {
        "start": r"\[\[(\w+)\s+type=([\w,\s\[\]]+)\]\]",
        "start_fstring": "[[{key} type={type}]]",
        "end_fstring": "[[/{key}]]",
        "whole": r"\[\[(\w+)\s+type=([\w,\s\[\]]+)\]\](.*?)\[\[/\1\]\]",
        "start_token": r"[",
        "end_token": r"]",
    }
    HTML = {
        "start": r"<(.*?)>",
        "start_fstring": "<{key} type={type}>",
        "end_fstring": "</{key}>",
        "whole": r"<(\w+)\s+type=([\w,\s\[\]]+)>(.*?)</\1>",  # also captures type
        "start_token": r"<",
        "end_token": r">",
    }


def get_pattern_by_type(parsing_type_value):
    return ParsingPattern[ParsingType(parsing_type_value).name].value



"""LLM for Development TestRun"""
import re
import os
import json
import openai

from typing import Any, AsyncGenerator, List, Dict, Optional, Union, Generator
from pydantic import BaseModel
from dotenv import load_dotenv
from litellm import acompletion

from promptmodel.utils.enums import ParsingType, ParsingPattern, get_pattern_by_type
from promptmodel.utils import logger
from promptmodel.utils.output_utils import convert_str_to_type
from promptmodel.utils.types import LLMStreamResponse

load_dotenv()


class OpenAIMessage(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = ""
    function_call: Optional[Dict[str, Any]] = None
    name: Optional[str] = None


class LLMDev:
    def __init__(self):
        self._model: str

    def __validate_openai_messages(
        self, messages: List[Dict[str, str]]
    ) -> List[OpenAIMessage]:
        """Validate and convert list of dictionaries to list of OpenAIMessage."""
        res = []
        for message in messages:
            res.append(OpenAIMessage(**message))
        return res

    async def dev_run(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: List[Any] = [],
        model: Optional[str] = None,
    ) -> AsyncGenerator[Any, None]:
        """Parse & stream output from openai chat completion."""
        _model = model or self._model
        raw_output = ""
        response = await acompletion(
            model=_model,
            messages=[
                message.model_dump(exclude_none=True)
                for message in self.__validate_openai_messages(messages)
            ],
            stream=True,
            functions=functions,
        )
        function_call = {"name": "", "arguments": ""}
        finish_reason_function_call = False
        async for chunk in response:
            if (
                "content" in chunk["choices"][0]["delta"]
                and chunk["choices"][0]["delta"]["content"] is not None
            ):
                stream_value = chunk["choices"][0]["delta"]["content"]
                raw_output += stream_value  # append raw output
                yield LLMStreamResponse(raw_output=stream_value)  # return raw output

            if (
                "function_call" in chunk["choices"][0]["delta"]
                and chunk["choices"][0]["delta"]["function_call"] is not None
            ):
                for key, value in chunk["choices"][0]["delta"]["function_call"].items():
                    function_call[key] += value

            if chunk["choices"][0]["finish_reason"] == "function_call":
                finish_reason_function_call = True
                yield LLMStreamResponse(function_call=function_call)

        # parsing
        if parsing_type and not finish_reason_function_call:
            parsing_pattern: Dict[str, str] = get_pattern_by_type(parsing_type)
            whole_pattern = parsing_pattern["whole"]
            parsed_results = re.findall(whole_pattern, raw_output, flags=re.DOTALL)
            for parsed_result in parsed_results:
                key = parsed_result[0]
                type_str = parsed_result[1]
                value = convert_str_to_type(parsed_result[2], type_str)
                yield LLMStreamResponse(parsed_outputs={key: value})

    async def dev_chat(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        model: Optional[str] = None,
    ) -> AsyncGenerator[Any, None]:
        """Parse & stream output from openai chat completion."""
        _model = model or self._model
        raw_output = ""
        response = await acompletion(
            model=_model,
            messages=[
                message.model_dump(exclude_none=True)
                for message in self.__validate_openai_messages(messages)
            ],
            stream=True,
        )
        async for chunk in response:
            if "content" in chunk["choices"][0]["delta"]:
                stream_value = chunk["choices"][0]["delta"]["content"]
                raw_output += stream_value  # append raw output
                yield LLMStreamResponse(raw_output=stream_value)  # return raw output

        # parsing
        if parsing_type:
            parsing_pattern: Dict[str, str] = get_pattern_by_type(parsing_type)
            whole_pattern = parsing_pattern["whole"]
            parsed_results = re.findall(whole_pattern, raw_output, flags=re.DOTALL)
            for parsed_result in parsed_results:
                key = parsed_result[0]
                type_str = parsed_result[1]
                value = convert_str_to_type(parsed_result[2], type_str)
                yield LLMStreamResponse(parsed_outputs={key: value})



import os
from dotenv import load_dotenv

load_dotenv()
testmode: str = os.environ.get("TESTMODE", "false")
if testmode == "true":
    # ENDPOINT_URL = "https://7015-211-193-21-129.ngrok-free.app/api/cli"
    # ENDPOINT_URL = (
    #     "https://promptmodel-fastapi-server-weavel-be-pr-107.up.railway.app/api/cli"
    # )
    ENDPOINT_URL = "https://promptmodel.up.railway.app/api/cli"
    # WEB_CLIENT_URL = "http://localhost:3000"
    WEB_CLIENT_URL = "https://app.promptmodel.run"
else:
    ENDPOINT_URL = "https://promptmodel.up.railway.app/api/cli"
    WEB_CLIENT_URL = "https://app.promptmodel.run"

GRANT_ACCESS_URL = "https://app.promptmodel.run/cli/grant-access"
PROMPTMODEL_DEV_FILENAME = os.path.join(os.getcwd(), "promptmodel_dev.py")
PROMPTMODEL_DEV_STARTER_FILENAME = "STARTER.py"



import time
from ...apis.base import APIClient
from requests import request
import typer
import webbrowser
from rich import print
from InquirerPy import inquirer

from ...constants import ENDPOINT_URL, WEB_CLIENT_URL
from ...utils.config_utils import read_config, upsert_config
from ...utils.crypto import generate_api_key, encrypt_message


def configure():
    """Saves user's default organization and project."""

    orgs = APIClient.execute(method="GET", path="/list_orgs").json()
    choices = [
        {
            "key": org["name"],
            "name": org["name"],
            "value": org,
        }
        for org in orgs
    ]
    org = inquirer.select(
        message="Select default organization:", choices=choices
    ).execute()

    projects = APIClient.execute(
        method="GET",
        path="/list_projects",
        params={"organization_id": org["organization_id"]},
    ).json()
    choices = [
        {
            "key": project["name"],
            "name": project["name"],
            "value": project,
        }
        for project in projects
    ]
    project = inquirer.select(
        message="Select default project:", choices=choices
    ).execute()

    upsert_config({"default_org": org, "default_project": project}, section="user")


app = typer.Typer(invoke_without_command=True, callback=configure)



# test.py
import openai

if __name__ == "__main__":
    # res = PromptModel("type_parsing").stream_and_parse()

    # for chunk in res:
    #     print(chunk)
    res = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        message=[
            { "role" : "system", "content" : "You are a helpfull assistant."},
            { "role" : "user", "content" : "hello?"}
        ]
    )
    print(res)


import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from promptmodel.llms.llm import LLM
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.utils.types import LLMResponse, LLMStreamResponse


def test_stream(mocker):
    llm = LLM()
    test_messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Introduce yourself in 50 words."},
    ]

    stream_res: Generator[LLMStreamResponse, None, None] = llm.stream(
        messages=test_messages,
        functions=[],
        model="gpt-3.5-turbo",
    )
    error_count = 0
    api_responses: List[ModelResponse] = []
    for res in stream_res:
        if res.error:
            error_count += 1
            print("ERROR")
            print(res.error)
            print(res.error_log)
        if res.api_response:
            api_responses.append(res.api_response)

    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"

    assert (
        api_responses[0].choices[0]["message"]["content"] is not None
    ), "content is None"
    assert api_responses[0]["response_ms"] is not None, "response_ms is None"

    # test logging
    llm_proxy = LLMProxy("test")

    mock_execute = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_execute.return_value = mock_response
    mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)

    llm_proxy._sync_log_to_cloud(
        version_uuid="test",
        inputs={},
        api_response=api_responses[0],
        parsed_outputs={},
        metadata={},
    )

    mock_execute.assert_called_once()
    _, kwargs = mock_execute.call_args

    assert (
        kwargs["json"]["api_response"] == api_responses[0].to_dict_recursive()
    ), "api_response is not equal"


def test_stream_with_functions(mocker):
    # JSON Schema to pass to OpenAI
    functions = [
        {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                },
                "required": ["location"],
            },
        }
    ]

    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    stream_res: Generator[LLMStreamResponse, None, None] = llm.stream(
        messages=messages,
        functions=functions,
        model="gpt-3.5-turbo-0613",
    )

    error_count = 0
    api_responses: List[ModelResponse] = []
    final_response: Optional[LLMStreamResponse] = None
    for res in stream_res:
        if res.error:
            error_count += 1
            print("ERROR")
            print(res.error)
            print(res.error_log)
        if res.api_response:
            api_responses.append(res.api_response)
            final_response = res

    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"

    assert (
        api_responses[0].choices[0]["message"]["content"] is None
    ), "content is not None"
    assert api_responses[0]["response_ms"] is not None, "response_ms is None"
    assert api_responses[0]["usage"] is not None, "usage is None"
    assert api_responses[0]["usage"]["prompt_tokens"] == 74, "prompt_tokens is not 74"

    # test logging
    llm_proxy = LLMProxy("test")

    mock_execute = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_execute.return_value = mock_response
    mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)

    llm_proxy._sync_log_to_cloud(
        version_uuid="test",
        inputs={},
        api_response=api_responses[0],
        parsed_outputs={},
        metadata={},
    )

    mock_execute.assert_called_once()
    _, kwargs = mock_execute.call_args

    assert (
        kwargs["json"]["api_response"] == api_responses[0].to_dict_recursive()
    ), "api_response is not equal"


@pytest.mark.asyncio
async def test_astream(mocker):
    llm = LLM()
    test_messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Introduce yourself in 50 words."},
    ]

    stream_res: Generator[LLMStreamResponse, None, None] = llm.astream(
        messages=test_messages,
        functions=[],
        model="gpt-3.5-turbo",
    )
    error_count = 0
    api_responses: List[ModelResponse] = []
    async for res in stream_res:
        if res.error:
            error_count += 1
            print("ERROR")
            print(res.error)
            print(res.error_log)
        if res.api_response:
            api_responses.append(res.api_response)

    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"

    assert (
        api_responses[0].choices[0]["message"]["content"] is not None
    ), "content is None"
    assert api_responses[0]["response_ms"] is not None, "response_ms is None"

    # test logging
    llm_proxy = LLMProxy("test")

    mock_execute = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_execute.return_value = mock_response
    mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)

    await llm_proxy._async_log_to_cloud(
        version_uuid="test",
        inputs={},
        api_response=api_responses[0],
        parsed_outputs={},
        metadata={},
    )

    mock_execute.assert_called_once()
    _, kwargs = mock_execute.call_args

    assert (
        kwargs["json"]["api_response"] == api_responses[0].to_dict_recursive()
    ), "api_response is not equal"


@pytest.mark.asyncio
async def test_astream_with_functions(mocker):
    # JSON Schema to pass to OpenAI
    functions = [
        {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                },
                "required": ["location"],
            },
        }
    ]

    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    stream_res: Generator[LLMStreamResponse, None, None] = llm.astream(
        messages=messages,
        functions=functions,
        model="gpt-3.5-turbo-0613",
    )

    error_count = 0
    api_responses: List[ModelResponse] = []
    final_response: Optional[LLMStreamResponse] = None
    async for res in stream_res:
        if res.error:
            error_count += 1
            print("ERROR")
            print(res.error)
            print(res.error_log)
        if res.api_response:
            api_responses.append(res.api_response)
            final_response = res

    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"

    assert (
        api_responses[0].choices[0]["message"]["content"] is None
    ), "content is not None"
    assert api_responses[0]["response_ms"] is not None, "response_ms is None"
    assert api_responses[0]["usage"] is not None, "usage is None"
    assert api_responses[0]["usage"]["prompt_tokens"] == 74, "prompt_tokens is not 74"

    # test logging
    llm_proxy = LLMProxy("test")

    mock_execute = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_execute.return_value = mock_response
    mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)

    await llm_proxy._async_log_to_cloud(
        version_uuid="test",
        inputs={},
        api_response=api_responses[0],
        parsed_outputs={},
        metadata={},
    )

    mock_execute.assert_called_once()
    _, kwargs = mock_execute.call_args

    assert (
        kwargs["json"]["api_response"] == api_responses[0].to_dict_recursive()
    ), "api_response is not equal"



import os
import sys
import typer
from promptmodel.cli.commands.login import app as login
from promptmodel.cli.commands.init import app as init
from promptmodel.cli.commands.dev import app as dev
from promptmodel.cli.commands.project import app as project
from promptmodel.cli.commands.configure import app as configure


# 현재 작업 디렉토리를 sys.path에 추가
current_working_directory = os.getcwd()
if current_working_directory not in sys.path:
    sys.path.append(current_working_directory)

app = typer.Typer(no_args_is_help=True)

app.add_typer(login, name="login")
app.add_typer(init, name="init")
app.add_typer(dev, name="dev")
app.add_typer(project, name="project")
app.add_typer(configure, name="configure")


if __name__ == "__main__":
    app()
